/* 
 *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <gtk/gtktable.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkstock.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <libgnome/gnome-i18n.h>
#include <string.h>

#include "ephy-bookmarks-editor.h"
#include "ephy-node-view.h"
#include "ephy-window.h"
#include "ephy-keywords-entry.h"
#include "ephy-dnd.h"

//#define DEBUG_MSG(x) g_print x
#define DEBUG_MSG(x)

static void ephy_bookmarks_editor_class_init (EphyBookmarksEditorClass *klass);
static void ephy_bookmarks_editor_init (EphyBookmarksEditor *editor);
static void ephy_bookmarks_editor_finalize (GObject *object);
static void ephy_bookmarks_editor_set_property (GObject *object,
		                                guint prop_id,
		                                const GValue *value,
		                                GParamSpec *pspec);
static void ephy_bookmarks_editor_get_property (GObject *object,
						guint prop_id,
						GValue *value,
						GParamSpec *pspec);

struct EphyBookmarksEditorPrivate
{
	EphyBookmarks *bookmarks;
	EphyNodeView *bm_view;
	EphyNodeView *key_view;
	EphyNodeFilter *bookmarks_filter;
	GtkWidget *title_entry;
	GtkWidget *keywords_entry;
	GtkWidget *search_entry;
};

enum
{
	PROP_0,
	PROP_BOOKMARKS
};

enum
{
	RESPONSE_REMOVE
};

static GObjectClass *parent_class = NULL;

GType
ephy_bookmarks_editor_get_type (void)
{
	static GType ephy_bookmarks_editor_type = 0;

	if (ephy_bookmarks_editor_type == 0)
	{
		static const GTypeInfo our_info =
		{
			sizeof (EphyBookmarksEditorClass),
			NULL,
			NULL,
			(GClassInitFunc) ephy_bookmarks_editor_class_init,
			NULL,
			NULL,
			sizeof (EphyBookmarksEditor),
			0,
			(GInstanceInitFunc) ephy_bookmarks_editor_init
		};

		ephy_bookmarks_editor_type = g_type_register_static (GTK_TYPE_DIALOG,
							             "EphyBookmarksEditor",
							             &our_info, 0);
	}

	return ephy_bookmarks_editor_type;
}

static void
ephy_bookmarks_editor_class_init (EphyBookmarksEditorClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	object_class->finalize = ephy_bookmarks_editor_finalize;

	object_class->set_property = ephy_bookmarks_editor_set_property;
	object_class->get_property = ephy_bookmarks_editor_get_property;

	g_object_class_install_property (object_class,
					 PROP_BOOKMARKS,
					 g_param_spec_object ("bookmarks",
							      "Bookmarks set",
							      "Bookmarks set",
							      EPHY_BOOKMARKS_TYPE,
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
}

static void
ephy_bookmarks_editor_finalize (GObject *object)
{
	EphyBookmarksEditor *editor;

	g_return_if_fail (object != NULL);
	g_return_if_fail (EPHY_IS_BOOKMARKS_EDITOR (object));

	editor = EPHY_BOOKMARKS_EDITOR (object);

	g_return_if_fail (editor->priv != NULL);

	g_object_unref (G_OBJECT (editor->priv->bookmarks_filter));

	g_free (editor->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
ephy_bookmarks_editor_node_selected_cb (GtkWidget *view,
				        EphyNode *node,
					EphyBookmarksEditor *editor)
{
	const char *title;
	const char *keywords;

	if (node != NULL)
	{
		g_assert (EPHY_IS_NODE (node));

		title = ephy_node_get_property_string
			(node, EPHY_NODE_BMK_PROP_TITLE);
		keywords = ephy_node_get_property_string
			(node, EPHY_NODE_BMK_PROP_KEYWORDS);
		gtk_entry_set_text (GTK_ENTRY (editor->priv->title_entry),
				    title ? g_strdup (title) : "");
		gtk_entry_set_text (GTK_ENTRY (editor->priv->keywords_entry),
				    keywords ? g_strdup (keywords) : "");
		gtk_widget_set_sensitive (GTK_WIDGET (editor->priv->title_entry), TRUE);
		gtk_widget_set_sensitive (GTK_WIDGET (editor->priv->keywords_entry), TRUE);
	}
	else
	{
		gtk_entry_set_text (GTK_ENTRY (editor->priv->title_entry), "");
		gtk_entry_set_text (GTK_ENTRY (editor->priv->keywords_entry), "");
		gtk_widget_set_sensitive (GTK_WIDGET (editor->priv->title_entry), FALSE);
		gtk_widget_set_sensitive (GTK_WIDGET (editor->priv->keywords_entry), FALSE);
	}
}

static void
ephy_bookmarks_editor_node_activated_cb (GtkWidget *view,
				         EphyNode *node,
					 EphyBookmarksEditor *editor)
{
	const char *location;
	GtkWindow *window;

	g_return_if_fail (EPHY_IS_NODE (node));
	location = ephy_node_get_property_string
		(node, EPHY_NODE_BMK_PROP_LOCATION);
	g_return_if_fail (location != NULL);

	window = gtk_window_get_transient_for (GTK_WINDOW (editor));
	g_return_if_fail (IS_EPHY_WINDOW (window));
	ephy_window_load_url (EPHY_WINDOW (window), location);
}

static void
ephy_bookmarks_editor_response_cb (GtkDialog *dialog,
		                   int response_id,
				   EphyBookmarksEditor *editor)
{
	switch (response_id)
	{
		case GTK_RESPONSE_CLOSE:
			gtk_widget_destroy (GTK_WIDGET (dialog));
			break;
		case RESPONSE_REMOVE:
			ephy_node_view_remove (editor->priv->bm_view);
			break;
	}
}

static void
update_prop_from_entry (EphyBookmarksEditor *editor,
		        GtkWidget *entry,
			guint id)
{
	GList *selection;
	GValue value = { 0, };

	selection = ephy_node_view_get_selection (editor->priv->bm_view);
	if (selection)
	{
		EphyNode *bm = EPHY_NODE (selection->data);
		char *tmp;

		tmp = gtk_editable_get_chars
			(GTK_EDITABLE (entry), 0, -1);
		g_value_init (&value, G_TYPE_STRING);
		g_value_set_string (&value, tmp);
		ephy_node_set_property (bm, id, &value);
	        g_value_unset (&value);
		g_free (tmp);
	}
	g_list_free (selection);
}

static void
title_entry_changed_cb (GtkWidget *entry, EphyBookmarksEditor *editor)
{
	update_prop_from_entry (editor, editor->priv->title_entry,
				EPHY_NODE_BMK_PROP_TITLE);
}

static void
keywords_changed_cb (GtkWidget *entry,
		     EphyBookmarksEditor *editor)
{
	EphyNode *node;
	GList *selection;
	char *keywords;

	selection = ephy_node_view_get_selection (editor->priv->bm_view);
	if (selection == NULL) return;
	node = EPHY_NODE (selection->data);
	g_list_free (selection);

	keywords = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
	ephy_bookmarks_update_keywords (editor->priv->bookmarks,
					keywords, node);
	g_free (keywords);

	update_prop_from_entry (editor, editor->priv->keywords_entry,
				EPHY_NODE_BMK_PROP_KEYWORDS);
}

static GtkWidget *
build_editing_table (EphyBookmarksEditor *editor)
{
	GtkWidget *table, *label, *entry;

	table = gtk_table_new (2, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);
	gtk_widget_show (table);

	label = gtk_label_new (NULL);
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_label_set_markup (GTK_LABEL (label), _("<b>Title:</b>"));
	gtk_widget_show (label);
	entry = gtk_entry_new ();
	editor->priv->title_entry = entry;
	gtk_widget_set_sensitive (GTK_WIDGET (entry), FALSE);
	gtk_widget_show (entry);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 0, 1);
	g_signal_connect (G_OBJECT (entry), "changed",
			  G_CALLBACK (title_entry_changed_cb),
			  editor);

	label = gtk_label_new (NULL);
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_label_set_markup (GTK_LABEL (label), _("<b>Keywords:</b>"));
	gtk_widget_show (label);
	entry = ephy_keywords_entry_new ();
	ephy_keywords_entry_set_bookmarks (EPHY_KEYWORDS_ENTRY (entry),
					   editor->priv->bookmarks);
	editor->priv->keywords_entry = entry;
	gtk_widget_set_sensitive (GTK_WIDGET (entry), FALSE);
	gtk_widget_show (entry);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 1, 2);
	g_signal_connect (G_OBJECT (entry), "keywords_changed",
			  G_CALLBACK (keywords_changed_cb),
			  editor);

	return table;
}

static void
bookmarks_filter (EphyBookmarksEditor *editor,
	          EphyNode *keyword)
{
	ephy_node_filter_empty (editor->priv->bookmarks_filter);
	ephy_node_filter_add_expression (editor->priv->bookmarks_filter,
				         ephy_node_filter_expression_new (EPHY_NODE_FILTER_EXPRESSION_HAS_PARENT,
								          keyword),
				         0);
	ephy_node_filter_done_changing (editor->priv->bookmarks_filter);
}

static void
keyword_node_selected_cb (EphyNodeView *view,
			  EphyNode *node,
			  EphyBookmarksEditor *editor)
{
	if (node == NULL)
	{
		ephy_node_view_select_node
			(editor->priv->key_view,
			 ephy_bookmarks_get_bookmarks
			 (editor->priv->bookmarks));
	}
	else
	{
		bookmarks_filter (editor, node);
	}
}

static void
search_entry_changed_cb (GtkWidget *entry, EphyBookmarksEditor *editor)
{
	char *search_text;

	search_text = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);

	GDK_THREADS_ENTER ();

	ephy_node_filter_empty (editor->priv->bookmarks_filter);
	ephy_node_filter_add_expression (editor->priv->bookmarks_filter,
				         ephy_node_filter_expression_new (EPHY_NODE_FILTER_EXPRESSION_STRING_PROP_CONTAINS,
								          EPHY_NODE_BMK_PROP_TITLE,
								          search_text),
				         0);
	ephy_node_filter_add_expression (editor->priv->bookmarks_filter,
				         ephy_node_filter_expression_new (EPHY_NODE_FILTER_EXPRESSION_STRING_PROP_CONTAINS,
								          EPHY_NODE_BMK_PROP_KEYWORDS,
								          search_text),
				         0);
	ephy_node_filter_done_changing (editor->priv->bookmarks_filter);

	GDK_THREADS_LEAVE ();

	g_free (search_text);
}

static GtkWidget *
build_search_box (EphyBookmarksEditor *editor)
{
	GtkWidget *box;
	GtkWidget *label;
	GtkWidget *entry;

	box = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (box);

	label = gtk_label_new (NULL);
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_label_set_markup (GTK_LABEL (label), _("<b>Search:</b>"));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (box),
			    label, FALSE, TRUE, 0);

	entry = gtk_entry_new ();
	editor->priv->search_entry = entry;
	gtk_widget_show (entry);
	gtk_box_pack_start (GTK_BOX (box),
			    entry, TRUE, TRUE, 0);
	g_signal_connect (G_OBJECT (entry), "changed",
			  G_CALLBACK (search_entry_changed_cb),
			  editor);
	return box;
}

static void
ephy_bookmarks_editor_construct (EphyBookmarksEditor *editor)
{
	GtkWidget *hbox, *vbox;
	EphyNodeView *bm_view, *key_view;
	EphyNode *node;

	gtk_dialog_set_has_separator (GTK_DIALOG (editor), FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (editor), 6);
	gtk_widget_set_size_request (GTK_WIDGET (editor), 500, 400);
	g_signal_connect (G_OBJECT (editor),
			  "response",
			  G_CALLBACK (ephy_bookmarks_editor_response_cb),
			  editor);

	hbox = gtk_hbox_new (FALSE, 6);
	gtk_container_set_border_width (GTK_CONTAINER (hbox), 6);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (editor)->vbox),
			    hbox, TRUE, TRUE, 0);
	gtk_widget_show (hbox);

	g_assert (editor->priv->bookmarks);

	node = ephy_bookmarks_get_keywords (editor->priv->bookmarks);
	key_view = ephy_node_view_new (node, NULL);
	ephy_node_view_set_browse_mode (key_view);
	ephy_node_view_add_column (key_view, _("Keywords"),
			           EPHY_TREE_MODEL_NODE_COL_KEYWORD, FALSE);
	gtk_box_pack_start (GTK_BOX (hbox), GTK_WIDGET (key_view), FALSE, TRUE, 0);
	gtk_widget_set_size_request (GTK_WIDGET (key_view), 120, -1);
	gtk_widget_show (GTK_WIDGET (key_view));
	editor->priv->key_view = key_view;
	g_signal_connect (G_OBJECT (key_view),
			  "node_selected",
			  G_CALLBACK (keyword_node_selected_cb),
			  editor);

	vbox = gtk_vbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX (hbox),
			    vbox, TRUE, TRUE, 0);
	gtk_widget_show (vbox);

	gtk_box_pack_start (GTK_BOX (vbox),
			    build_search_box (editor),
			    FALSE, FALSE, 0);

	node = ephy_bookmarks_get_bookmarks (editor->priv->bookmarks);
	editor->priv->bookmarks_filter = ephy_node_filter_new ();
	bm_view = ephy_node_view_new (node, editor->priv->bookmarks_filter);
	ephy_node_view_enable_drag_source (bm_view);
	ephy_node_view_add_column (bm_view, _("Title"),
				   EPHY_TREE_MODEL_NODE_COL_BOOKMARK, TRUE);
	ephy_node_view_add_column (bm_view, _("Location"),
				   EPHY_TREE_MODEL_NODE_COL_LOCATION, TRUE);
	gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (bm_view), TRUE, TRUE, 0);
	gtk_widget_show (GTK_WIDGET (bm_view));
	editor->priv->bm_view = bm_view;
	g_signal_connect (G_OBJECT (bm_view),
			  "node_activated",
			  G_CALLBACK (ephy_bookmarks_editor_node_activated_cb),
			  editor);
	g_signal_connect (G_OBJECT (bm_view),
			  "node_selected",
			  G_CALLBACK (ephy_bookmarks_editor_node_selected_cb),
			  editor);

	gtk_box_pack_start (GTK_BOX (vbox),
			    build_editing_table (editor),
			    FALSE, FALSE, 0);

	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_REMOVE,
			       RESPONSE_REMOVE);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_CLOSE,
			       GTK_RESPONSE_CLOSE);
	gtk_dialog_set_default_response (GTK_DIALOG (editor), GTK_RESPONSE_CLOSE);
}

GtkWidget *
ephy_bookmarks_editor_new (EphyBookmarks *bookmarks,
			   GtkWindow *parent)
{
	EphyBookmarksEditor *editor;

	g_assert (bookmarks != NULL);

	editor = EPHY_BOOKMARKS_EDITOR (g_object_new
			(EPHY_TYPE_BOOKMARKS_EDITOR,
			 "bookmarks", bookmarks,
			 NULL));

	if (parent)
	{
		gtk_window_set_transient_for (GTK_WINDOW (editor), parent);
	}

	ephy_bookmarks_editor_construct (editor);

	return GTK_WIDGET (editor);
}

static void
ephy_bookmarks_editor_set_property (GObject *object,
		                    guint prop_id,
		                    const GValue *value,
		                    GParamSpec *pspec)
{
	EphyBookmarksEditor *editor = EPHY_BOOKMARKS_EDITOR (object);

	switch (prop_id)
	{
	case PROP_BOOKMARKS:
		editor->priv->bookmarks = g_value_get_object (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
ephy_bookmarks_editor_get_property (GObject *object,
		                    guint prop_id,
		                    GValue *value,
		                    GParamSpec *pspec)
{
	EphyBookmarksEditor *editor = EPHY_BOOKMARKS_EDITOR (object);

	switch (prop_id)
	{
	case PROP_BOOKMARKS:
		g_value_set_object (value, editor->priv->bookmarks);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
ephy_bookmarks_editor_init (EphyBookmarksEditor *editor)
{
	editor->priv = g_new0 (EphyBookmarksEditorPrivate, 1);
}
